@@ -1,3 +1,4 @@
+
 from parsimonious.utils import StrAndRepr
 
 
@@ -7,14 +8,14 @@ class ParseError(StrAndRepr, Exception):
     def __init__(self, text, pos=-1, expr=None):
         # It would be nice to use self.args, but I don't want to pay a penalty
         # to call descriptors or have the confusion of numerical indices in
-        # Expression._match().
+        # Expression.match_core().
         self.text = text
         self.pos = pos
         self.expr = expr
 
-    def __unicode__(self):
+    def __str__(self):
         rule_name = ((u"'%s'" % self.expr.name) if self.expr.name else
-                     unicode(self.expr))
+                     str(self.expr))
         return u"Rule %s didn't match at '%s' (line %s, column %s)." % (
                 rule_name,
                 self.text[self.pos:self.pos + 20],
@@ -44,7 +45,7 @@ class IncompleteParseError(ParseError):
     """A call to ``parse()`` matched a whole Expression but did not consume the
     entire text."""
 
-    def __unicode__(self):
+    def __str__(self):
         return u"Rule '%s' matched in its entirety, but it didn't consume all the text. The non-matching portion of the text begins with '%s' (line %s, column %s)." % (
                 self.expr.name,
                 self.text[self.pos:self.pos + 20],
@@ -81,7 +82,16 @@ class VisitationError(Exception):
              node.prettily(error=node)))
 
 
-class UndefinedLabel(StrAndRepr, VisitationError):
+class BadGrammar(StrAndRepr, Exception):
+    """Something was wrong with the definition of a grammar.
+
+    Note that a ParseError might be raised instead if the error is in the
+    grammar definition syntax.
+
+    """
+
+
+class UndefinedLabel(BadGrammar):
     """A rule referenced in a grammar was never defined.
 
     Circular references and forward references are okay, but you have to define
@@ -91,5 +101,5 @@ class UndefinedLabel(StrAndRepr, VisitationError):
     def __init__(self, label):
         self.label = label
 
-    def __unicode__(self):
+    def __str__(self):
         return u'The label "%s" was never defined.' % self.label
