@@ -6,17 +6,21 @@ These do the parsing.
 # TODO: Make sure all symbol refs are local--not class lookups or
 # anything--for speed. And kill all the dots.
 
-from inspect import getargspec
-import re
+from inspect import getfullargspec, isfunction, ismethod, ismethoddescriptor
+import regex as re
 
 from parsimonious.exceptions import ParseError, IncompleteParseError
 from parsimonious.nodes import Node, RegexNode
 from parsimonious.utils import StrAndRepr
 
-
 MARKER = object()
 
 
+def is_callable(value):
+    criteria = [isfunction, ismethod, ismethoddescriptor]
+    return any([criterion(value) for criterion in criteria])
+
+
 def expression(callable, rule_name, grammar):
     """Turn a plain callable into an Expression.
 
@@ -55,7 +59,16 @@ def expression(callable, rule_name, grammar):
         part of, to make delegating to other rules possible
 
     """
-    num_args = len(getargspec(callable).args)
+
+    # Resolve unbound methods; allows grammars to use @staticmethod custom rules
+    # https://stackoverflow.com/questions/41921255/staticmethod-object-is-not-callable
+    if ismethoddescriptor(callable) and hasattr(callable, '__func__'):
+        callable = callable.__func__
+
+    num_args = len(getfullargspec(callable).args)
+    if ismethod(callable):
+        # do not count the first argument (typically 'self') for methods
+        num_args -= 1
     if num_args == 2:
         is_simple = True
     elif num_args == 5:
@@ -69,14 +82,14 @@ def expression(callable, rule_name, grammar):
             result = (callable(text, pos) if is_simple else
                       callable(text, pos, cache, error, grammar))
 
-            if isinstance(result, (long, int)):
+            if isinstance(result, int):
                 end, children = result, None
             elif isinstance(result, tuple):
                 end, children = result
             else:
                 # Node or None
                 return result
-            return Node(self.name, text, pos, end, children=children)
+            return Node(self, text, pos, end, children=children)
 
         def _as_rhs(self):
             return '{custom function "%s"}' % callable.__name__
@@ -91,10 +104,20 @@ class Expression(StrAndRepr):
     # http://stackoverflow.com/questions/1336791/dictionary-vs-object-which-is-more-efficient-and-why
 
     # Top-level expressions--rules--have names. Subexpressions are named ''.
-    __slots__ = ['name']
+    __slots__ = ['name', 'identity_tuple']
 
     def __init__(self, name=''):
         self.name = name
+        self.identity_tuple = (self.name, )
+
+    def __hash__(self):
+        return hash(self.identity_tuple)
+
+    def __eq__(self, other):
+        return isinstance(other, self.__class__) and self.identity_tuple == other.identity_tuple
+
+    def __ne__(self, other):
+        return not (self == other)
 
     def parse(self, text, pos=0):
         """Return a parse tree of ``text``.
@@ -170,17 +193,16 @@ class Expression(StrAndRepr):
             # Don't bother reporting on unnamed expressions (unless that's all
             # we've seen so far), as they're hard to track down for a human.
             # Perhaps we could include the unnamed subexpressions later as
-            # auxilliary info.
+            # auxiliary info.
             error.expr = self
             error.pos = pos
 
         return node
 
-    def __unicode__(self):
-        return u'<%s %s at 0x%s>' % (
+    def __str__(self):
+        return u'<%s %s>' % (
             self.__class__.__name__,
-            self.as_rule(),
-            id(self))
+            self.as_rule())
 
     def as_rule(self):
         """Return the left- and right-hand sides of a rule that represents me.
@@ -188,8 +210,11 @@ class Expression(StrAndRepr):
         Return unicode. If I have no ``name``, omit the left-hand side.
 
         """
-        return ((u'%s = %s' % (self.name, self._as_rhs())) if self.name else
-                self._as_rhs())
+        rhs = self._as_rhs().strip()
+        if rhs.startswith('(') and rhs.endswith(')'):
+            rhs = rhs[1:-1]
+
+        return (u'%s = %s' % (self.name, rhs)) if self.name else rhs
 
     def _unicode_members(self):
         """Return an iterable of my unicode-represented children, stopping
@@ -217,14 +242,25 @@ class Literal(Expression):
     def __init__(self, literal, name=''):
         super(Literal, self).__init__(name)
         self.literal = literal
+        self.identity_tuple = (name, literal)
 
     def _uncached_match(self, text, pos, cache, error):
         if text.startswith(self.literal, pos):
-            return Node(self.name, text, pos, pos + len(self.literal))
+            return Node(self, text, pos, pos + len(self.literal))
 
     def _as_rhs(self):
-        # TODO: Get backslash escaping right.
-        return '"%s"' % self.literal
+        return repr(self.literal)
+
+
+class TokenMatcher(Literal):
+    """An expression matching a single token of a given type
+
+    This is for use only with TokenGrammars.
+
+    """
+    def _uncached_match(self, token_list, pos, cache, error):
+        if token_list[pos].type == self.literal:
+            return Node(self, token_list, pos, pos + 1)
 
 
 class Regex(Expression):
@@ -237,33 +273,34 @@ class Regex(Expression):
     __slots__ = ['re']
 
     def __init__(self, pattern, name='', ignore_case=False, locale=False,
-                 multiline=False, dot_all=False, unicode=False, verbose=False):
+                 multiline=False, dot_all=False, unicode=False, verbose=False, ascii=False):
         super(Regex, self).__init__(name)
         self.re = re.compile(pattern, (ignore_case and re.I) |
                                       (locale and re.L) |
                                       (multiline and re.M) |
                                       (dot_all and re.S) |
                                       (unicode and re.U) |
-                                      (verbose and re.X))
+                                      (verbose and re.X) |
+                                      (ascii and re.A))
+        self.identity_tuple = (self.name, self.re)
 
     def _uncached_match(self, text, pos, cache, error):
         """Return length of match, ``None`` if no match."""
         m = self.re.match(text, pos)
         if m is not None:
             span = m.span()
-            node = RegexNode(self.name, text, pos, pos + span[1] - span[0])
+            node = RegexNode(self, text, pos, pos + span[1] - span[0])
             node.match = m  # TODO: A terrible idea for cache size?
             return node
 
     def _regex_flags_from_bits(self, bits):
         """Return the textual equivalent of numerically encoded regex flags."""
-        flags = 'tilmsux'
-        return ''.join(flags[i] if (1 << i) & bits else '' for i in xrange(6))
+        flags = 'ilmsuxa'
+        return ''.join(flags[i - 1] if (1 << i) & bits else '' for i in range(1, len(flags) + 1))
 
     def _as_rhs(self):
-        # TODO: Get backslash escaping right.
-        return '~"%s"%s' % (self.re.pattern,
-                            self._regex_flags_from_bits(self.re.flags))
+        return '~{!r}{}'.format(self.re.pattern,
+                                self._regex_flags_from_bits(self.re.flags))
 
 
 class Compound(Expression):
@@ -276,6 +313,18 @@ class Compound(Expression):
         super(Compound, self).__init__(kwargs.get('name', ''))
         self.members = members
 
+    def __hash__(self):
+        # Note we leave members out of the hash computation, since compounds can get added to
+        # sets, then have their members mutated. See RuleVisitor._resolve_refs.
+        # Equality should still work, but we want the rules to go into the correct hash bucket.
+        return hash((self.__class__, self.name))
+
+    def __eq__(self, other):
+        return (
+            isinstance(other, self.__class__) and
+            self.name == other.name and
+            self.members == other.members)
+
 
 class Sequence(Compound):
     """A series of expressions that must match contiguous, ordered pieces of
@@ -298,10 +347,11 @@ class Sequence(Compound):
             new_pos += length
             length_of_sequence += length
         # Hooray! We got through all the members!
-        return Node(self.name, text, pos, pos + length_of_sequence, children)
+        return Node(self, text, pos, pos + length_of_sequence, children)
 
     def _as_rhs(self):
-        return u' '.join(self._unicode_members())
+        return u'({0})'.format(u' '.join(self._unicode_members()))
+
 
 class OneOf(Compound):
     """A series of expressions, one of which must match
@@ -315,10 +365,10 @@ class OneOf(Compound):
             node = m.match_core(text, pos, cache, error)
             if node is not None:
                 # Wrap the succeeding child in a node representing the OneOf:
-                return Node(self.name, text, pos, node.end, children=[node])
+                return Node(self, text, pos, node.end, children=[node])
 
     def _as_rhs(self):
-        return u' / '.join(self._unicode_members())
+        return u'({0})'.format(u' / '.join(self._unicode_members()))
 
 
 class Lookahead(Compound):
@@ -332,7 +382,7 @@ class Lookahead(Compound):
     def _uncached_match(self, text, pos, cache, error):
         node = self.members[0].match_core(text, pos, cache, error)
         if node is not None:
-            return Node(self.name, text, pos, pos)
+            return Node(self, text, pos, pos)
 
     def _as_rhs(self):
         return u'&%s' % self._unicode_members()[0]
@@ -349,7 +399,7 @@ class Not(Compound):
         # not bother to cache NOTs directly.
         node = self.members[0].match_core(text, pos, cache, error)
         if node is None:
-            return Node(self.name, text, pos, pos)
+            return Node(self, text, pos, pos)
 
     def _as_rhs(self):
         # TODO: Make sure this parenthesizes the member properly if it's an OR
@@ -368,8 +418,8 @@ class Optional(Compound):
     """
     def _uncached_match(self, text, pos, cache, error):
         node = self.members[0].match_core(text, pos, cache, error)
-        return (Node(self.name, text, pos, pos) if node is None else
-                Node(self.name, text, pos, node.end, children=[node]))
+        return (Node(self, text, pos, pos) if node is None else
+                Node(self, text, pos, node.end, children=[node]))
 
     def _as_rhs(self):
         return u'%s?' % self._unicode_members()[0]
@@ -378,6 +428,7 @@ class Optional(Compound):
 # TODO: Merge with OneOrMore.
 class ZeroOrMore(Compound):
     """An expression wrapper like the * quantifier in regexes."""
+
     def _uncached_match(self, text, pos, cache, error):
         new_pos = pos
         children = []
@@ -385,7 +436,7 @@ class ZeroOrMore(Compound):
             node = self.members[0].match_core(text, new_pos, cache, error)
             if node is None or not (node.end - node.start):
                 # Node was None or 0 length. 0 would otherwise loop infinitely.
-                return Node(self.name, text, pos, new_pos, children)
+                return Node(self, text, pos, new_pos, children)
             children.append(node)
             new_pos += node.end - node.start
 
@@ -422,7 +473,7 @@ class OneOrMore(Compound):
                 break
             new_pos += length
         if len(children) >= self.min:
-            return Node(self.name, text, pos, new_pos, children)
+            return Node(self, text, pos, new_pos, children)
 
     def _as_rhs(self):
         return u'%s+' % self._unicode_members()[0]
