@@ -5,17 +5,16 @@ optimizations that would be tedious to do when constructing an expression tree
 by hand.
 
 """
-import ast
-from inspect import isfunction, ismethod
+from collections import OrderedDict
 
-from parsimonious.exceptions import UndefinedLabel
+from parsimonious.exceptions import BadGrammar, UndefinedLabel
 from parsimonious.expressions import (Literal, Regex, Sequence, OneOf,
-    Lookahead, Optional, ZeroOrMore, OneOrMore, Not, expression)
+    Lookahead, Optional, ZeroOrMore, OneOrMore, Not, TokenMatcher,
+    expression, is_callable)
 from parsimonious.nodes import NodeVisitor
-from parsimonious.utils import StrAndRepr
+from parsimonious.utils import evaluate_string
 
-
-class Grammar(StrAndRepr, dict):
+class Grammar(OrderedDict):
     """A collection of rules that describe a language
 
     You can start parsing from the default rule by calling ``parse()``
@@ -58,14 +57,13 @@ class Grammar(StrAndRepr, dict):
             ``rules`` in case of naming conflicts.
 
         """
-        decorated_custom_rules = dict(
-            (k, expression(v, k, self) if isfunction(v) or
-                                          ismethod(v) else
-                v) for k, v in more_rules.iteritems())
+
+        decorated_custom_rules = {
+            k: (expression(v, k, self) if is_callable(v) else v)
+            for k, v in more_rules.items()}
 
         exprs, first = self._expressions_from_rules(rules, decorated_custom_rules)
-
-        self.update(exprs)
+        super(Grammar, self).__init__(exprs.items())
         self.default_rule = first  # may be None
 
     def default(self, rule_name):
@@ -82,7 +80,8 @@ class Grammar(StrAndRepr, dict):
         no Expressions.
 
         """
-        new = Grammar(**self)
+        new = Grammar.__new__(Grammar)
+        super(Grammar, new).__init__(self.items())
         new.default_rule = self.default_rule
         return new
 
@@ -128,17 +127,29 @@ class Grammar(StrAndRepr, dict):
                                "default rule. Choose a specific rule instead, "
                                "like some_grammar['some_rule'].parse(...).")
 
-    def __unicode__(self):
+    def __str__(self):
         """Return a rule string that, when passed to the constructor, would
         reconstitute the grammar."""
         exprs = [self.default_rule] if self.default_rule else []
-        exprs.extend(expr for expr in self.itervalues() if
+        exprs.extend(expr for expr in self.values() if
                      expr is not self.default_rule)
         return '\n'.join(expr.as_rule() for expr in exprs)
 
     def __repr__(self):
         """Return an expression that will reconstitute the grammar."""
-        return "Grammar('%s')" % str(self).encode('string_escape')
+        return "Grammar({!r})".format(str(self))
+
+
+class TokenGrammar(Grammar):
+    """A Grammar which takes a list of pre-lexed tokens instead of text
+
+    This is useful if you want to do the lexing yourself, as a separate pass:
+    for example, to implement indentation-based languages.
+
+    """
+    def _expressions_from_rules(self, rules, custom_rules):
+        tree = rule_grammar.parse(rules)
+        return TokenRuleVisitor(custom_rules).visit(tree)
 
 
 class BootstrappingGrammar(Grammar):
@@ -174,7 +185,7 @@ class BootstrappingGrammar(Grammar):
         literal = Sequence(spaceless_literal, _, name='literal')
         regex = Sequence(Literal('~'),
                          literal,
-                         Regex('[ilmsux]*', ignore_case=True),
+                         Regex('[ilmsuxa]*', ignore_case=True),
                          _,
                          name='regex')
         atom = OneOf(reference, literal, regex, name='atom')
@@ -227,7 +238,7 @@ rule_syntax = (r'''
     term = not_term / lookahead_term / quantified / atom
     quantified = atom quantifier
     atom = reference / literal / regex / parenthesized
-    regex = "~" spaceless_literal ~"[ilmsux]*"i _
+    regex = "~" spaceless_literal ~"[ilmsuxa]*"i _
     parenthesized = "(" _ expression ")" _
     quantifier = ~"[*+?]" _
     reference = label !equals
@@ -244,7 +255,7 @@ rule_syntax = (r'''
     ''')
 
 
-class LazyReference(unicode):
+class LazyReference(str):
     """A lazy reference to a rule, which we resolve after grokking all the
     rules"""
 
@@ -276,64 +287,74 @@ class RuleVisitor(NodeVisitor):
         """
         self.custom_rules = custom_rules or {}
 
-    def visit_parenthesized(self, parenthesized, (left_paren, _1,
-                                                  expression,
-                                                  right_paren, _2)):
+    def visit_parenthesized(self, node, parenthesized):
         """Treat a parenthesized subexpression as just its contents.
 
         Its position in the tree suffices to maintain its grouping semantics.
 
         """
+        left_paren, _, expression, right_paren, _ = parenthesized
         return expression
 
-    def visit_quantifier(self, quantifier, (symbol, _)):
+    def visit_quantifier(self, node, quantifier):
         """Turn a quantifier into just its symbol-matching node."""
+        symbol, _ = quantifier
         return symbol
 
-    def visit_quantified(self, quantified, (atom, quantifier)):
+    def visit_quantified(self, node, quantified):
+        atom, quantifier = quantified
         return self.quantifier_classes[quantifier.text](atom)
 
-    def visit_lookahead_term(self, lookahead_term, (ampersand, term, _)):
+    def visit_lookahead_term(self, node, lookahead_term):
+        ampersand, term, _ = lookahead_term
         return Lookahead(term)
 
-    def visit_not_term(self, not_term, (exclamation, term, _)):
+    def visit_not_term(self, node, not_term):
+        exclamation, term, _ = not_term
         return Not(term)
 
-    def visit_rule(self, rule, (label, equals, expression)):
+    def visit_rule(self, node, rule):
         """Assign a name to the Expression and return it."""
+        label, equals, expression = rule
         expression.name = label  # Assign a name to the expr.
         return expression
 
-    def visit_sequence(self, sequence, (term, other_terms)):
+    def visit_sequence(self, node, sequence):
         """A parsed Sequence looks like [term node, OneOrMore node of
         ``another_term``s]. Flatten it out."""
+        term, other_terms = sequence
         return Sequence(term, *other_terms)
 
-    def visit_ored(self, ored, (first_term, other_terms)):
+    def visit_ored(self, node, ored):
+        first_term, other_terms = ored
         return OneOf(first_term, *other_terms)
 
-    def visit_or_term(self, or_term, (slash, _, term)):
+    def visit_or_term(self, node, or_term):
         """Return just the term from an ``or_term``.
 
         We already know it's going to be ored, from the containing ``ored``.
 
         """
+        slash, _, term = or_term
         return term
 
-    def visit_label(self, label, (name, _)):
+    def visit_label(self, node, label):
         """Turn a label into a unicode string."""
+        name, _ = label
         return name.text
 
-    def visit_reference(self, reference, (label, not_equals)):
+    def visit_reference(self, node, reference):
         """Stick a :class:`LazyReference` in the tree as a placeholder.
 
         We resolve them all later.
 
         """
+        label, not_equals = reference
         return LazyReference(label)
 
-    def visit_regex(self, regex, (tilde, literal, flags, _)):
+    def visit_regex(self, node, regex):
         """Return a ``Regex`` expression."""
+        tilde, literal, flags, _ = regex
         flags = flags.text.upper()
         pattern = literal.literal  # Pull the string back out of the Literal
                                    # object.
@@ -342,18 +363,16 @@ class RuleVisitor(NodeVisitor):
                               multiline='M' in flags,
                               dot_all='S' in flags,
                               unicode='U' in flags,
-                              verbose='X' in flags)
+                              verbose='X' in flags,
+                              ascii='A' in flags)
 
     def visit_spaceless_literal(self, spaceless_literal, visited_children):
         """Turn a string literal into a ``Literal`` that recognizes it."""
-        # Piggyback on Python's string support so we can have backslash
-        # escaping and niceties like \n, \t, etc.
-        # string.decode('string_escape') would have been a lower-level
-        # possibility.
-        return Literal(ast.literal_eval(spaceless_literal.text))
+        return Literal(evaluate_string(spaceless_literal.text))
 
-    def visit_literal(self, literal, (spaceless_literal, _)):
+    def visit_literal(self, node, literal):
         """Pick just the literal out of a literal-and-junk combo."""
+        spaceless_literal, _ = literal
         return spaceless_literal
 
     def generic_visit(self, node, visited_children):
@@ -385,7 +404,7 @@ class RuleVisitor(NodeVisitor):
 
         """
         if isinstance(expr, LazyReference):
-            label = unicode(expr)
+            label = str(expr)
             try:
                 reffed_expr = rule_map[label]
             except KeyError:
@@ -397,11 +416,11 @@ class RuleVisitor(NodeVisitor):
                 # of `expr.members` can refer back to `expr`, but it can't go
                 # any farther.
                 done.add(expr)
-                expr.members = [self._resolve_refs(rule_map, member, done)
-                                for member in expr.members]
+                expr.members = tuple(self._resolve_refs(rule_map, member, done)
+                                     for member in expr.members)
             return expr
 
-    def visit_rules(self, node, (_, rules)):
+    def visit_rules(self, node, rules_list):
         """Collate all the rules into a map. Return (map, default rule).
 
         The default rule is the first one. Or, if you have more than one rule
@@ -411,11 +430,13 @@ class RuleVisitor(NodeVisitor):
         due to being kwarg-based, are unordered.
 
         """
+        _, rules = rules_list
+
         # Map each rule's name to its Expression. Later rules of the same name
         # override earlier ones. This lets us define rules multiple times and
         # have the last declaration win, so you can extend grammars by
         # concatenation.
-        rule_map = dict((expr.name, expr) for expr in rules)
+        rule_map = OrderedDict((expr.name, expr) for expr in rules)
 
         # And custom rules override string-based rules. This is the least
         # surprising choice when you compare the dict constructor:
@@ -424,8 +445,8 @@ class RuleVisitor(NodeVisitor):
 
         # Resolve references. This tolerates forward references.
         done = set()
-        rule_map = dict((expr.name, self._resolve_refs(rule_map, expr, done))
-                        for expr in rule_map.itervalues())
+        rule_map = OrderedDict((expr.name, self._resolve_refs(rule_map, expr, done))
+                               for expr in rule_map.values())
 
         # isinstance() is a temporary hack around the fact that * rules don't
         # always get transformed into lists by NodeVisitor. We should fix that;
@@ -434,6 +455,22 @@ class RuleVisitor(NodeVisitor):
                           if isinstance(rules, list) and rules else None)
 
 
+class TokenRuleVisitor(RuleVisitor):
+    """A visitor which builds expression trees meant to work on sequences of
+    pre-lexed tokens rather than strings"""
+
+    def visit_spaceless_literal(self, spaceless_literal, visited_children):
+        """Turn a string literal into a ``TokenMatcher`` that matches
+        ``Token`` objects by their ``type`` attributes."""
+        return TokenMatcher(evaluate_string(spaceless_literal.text))
+
+    def visit_regex(self, node, regex):
+        tilde, literal, flags, _ = regex
+        raise BadGrammar('Regexes do not make sense in TokenGrammars, since '
+                         'TokenGrammars operate on pre-lexed tokens rather '
+                         'than characters.')
+
+
 # Bootstrap to level 1...
 rule_grammar = BootstrappingGrammar(rule_syntax)
 # ...and then to level 2. This establishes that the node tree of our rule
